using System;
using System.Collections.Generic;
using System.Text;
using SideShow;
using SideShow.Channel;
using System.Windows.Forms;
using SideShow.Scf.Events;
using SideShow.Scf.ObjectModel;
using SideShow.Contents;
using System.Drawing;
using PocketPCSideShowClient;
using SideShow.Events;

namespace SideShowClient
{
    public class ScfNavigator
    {
        private SideShowApplication _application;
        private SideShowEndpoint _endPoint;
        private IClientChannel _channel;
        private SideShow.Scf.Contents.ScfContent _currentPage;
        private SideShow.Scf.Contents.ScfContent _previousPage;
        private Control _container;
        private Dictionary<SideShow.Scf.Contents.ScfContent, ScfUserControl> _pages;
        private Stack<SideShow.Scf.Contents.ScfContent> _backStack;

        public event EventHandler CurrentPageChanged;
        public event EventHandler GotoHome;

        public ScfNavigator(Control container, SideShowApplication app, IClientChannel channel)
        {
            _backStack = new Stack<SideShow.Scf.Contents.ScfContent>();
            _pages = new Dictionary<SideShow.Scf.Contents.ScfContent, ScfUserControl>();
            _container = container;
            _application = app;
            _endPoint = app.Endpoints[SideShowEndpoint.ScfFormatId];
            _endPoint.ContentAdded += new EventHandler<ContentEventArgs>(EndPointContentAdded);
            _channel = channel;
            Navigate(1);
        }

        private void EndPointContentAdded(object sender, ContentEventArgs e)
        {
            if (_currentPage != null && e.Content.Id == _currentPage.Id)
            {
                _container.Invoke(new Action<SideShow.Scf.Contents.ScfContent>(Update), (SideShow.Scf.Contents.ScfContent)e.Content);
            }
        }

        private void Update(SideShow.Scf.Contents.ScfContent content)
        {
            ScfUserControl ctrl = (ScfUserControl)_container.Controls[0];
            ctrl.ScfObject = content.ScfDocument.Body;
        }

        private void OnCurrentPageChanged()
        {
            if (CurrentPageChanged != null)
            {
                CurrentPageChanged(this, EventArgs.Empty);
            }
        }

        private void OnGotoHome()
        {
            if (GotoHome != null)
            {
                GotoHome(this, EventArgs.Empty);
            }
        }

        public void Home()
        {
            OnGotoHome();
        }

        public string Title
        {
            get { return _currentPage.ScfDocument.Body.Title; }
        }

        public SideShow.Scf.Contents.ScfContent CurrentPage
        {
            get { return _currentPage; }
        }

        public Image GetImage(uint id)
        {
            ImageContent ic = (ImageContent) _endPoint.GetContent(id);
            return ic.Image;
        }

        public Content GetContent(uint id)
        {
            return _endPoint.GetContent(id);
        }

        public void Back()
        {
            if (_backStack.Count > 0)
            {
                _previousPage = _currentPage;
                _currentPage = _backStack.Pop();
                uint prevId = _previousPage == null ? 0 : _previousPage.Id;
                ScfNavigationEvent e = new ScfNavigationEvent(_application.Id, _endPoint.Id, prevId, _currentPage.Id, SideShow.Scf.ScfButton.Back);
                _channel.PostEvent(e);
                UserControl ctrl = GetPage(_currentPage);
                _container.Controls.Clear();
                _container.Controls.Add(ctrl);
                ctrl.Focus();
                OnCurrentPageChanged();
            }
            else
            {
                OnGotoHome();
            }
        }

        public void Navigate(uint target)
        {
            Navigate(target, SideShow.Scf.ScfButton.Select);
        }

        public void Navigate(uint target, SideShow.Scf.ScfButton btn)
        {
            _previousPage = _currentPage;
            uint prevId = _previousPage == null ? 0 : _previousPage.Id;
            SideShow.Scf.Contents.ScfContent t = (SideShow.Scf.Contents.ScfContent)_endPoint.GetContent(target);
            _currentPage = t ?? _currentPage;
            if (_currentPage == null)
            {
                ContentMissingEvent cme = new ContentMissingEvent(_application.Id, _endPoint.Id, target);
                _channel.PostEvent(cme);
            }
            ScfNavigationEvent e = new ScfNavigationEvent(_application.Id, _endPoint.Id, prevId, target, btn);
            _channel.PostEvent(e);
            if (_previousPage != null && _previousPage.ScfDocument.Body is ScfMenu)
            {
                if (_backStack.Count == 0 || (_backStack.Peek() != _previousPage && _previousPage != _currentPage))
                {
                    _backStack.Push(_previousPage);
                }
            }
            UserControl ctrl = GetPage(_currentPage);
            _container.Controls.Clear();
            _container.Controls.Add(ctrl);
            ctrl.Focus();
            OnCurrentPageChanged();
        }

        

        public void Navigate(ScfMenuItem menuItem)
        {
            _previousPage = _currentPage;
            uint prevId = _previousPage == null ? 0 : _previousPage.Id;
            SideShow.Scf.Contents.ScfContent t = (SideShow.Scf.Contents.ScfContent) _endPoint.GetContent(menuItem.TargetId);
            _currentPage = t ?? _currentPage;
            if (_currentPage == null)
            {
                ContentMissingEvent cme = new ContentMissingEvent(_application.Id, _endPoint.Id, menuItem.TargetId);
                _channel.PostEvent(cme);
            }
            ScfMenuActionEvent e = new ScfMenuActionEvent(_application.Id, _endPoint.Id, prevId, menuItem.TargetId, SideShow.Scf.ScfButton.Select, menuItem.Id); 
            _channel.PostEvent(e);
            if (_previousPage != null && _previousPage.ScfDocument.Body is ScfMenu)
            {
                if (_backStack.Count == 0 || (_backStack.Peek() != _previousPage && _previousPage != _currentPage))
                {
                    _backStack.Push(_previousPage);
                }
            }
            UserControl ctrl = GetPage(_currentPage);
            _container.Controls.Clear();
            _container.Controls.Add(ctrl);
            ctrl.Focus();
            OnCurrentPageChanged();
        }

        public void Navigate(ScfMenuItem menuItem, ScfMenu ctxMenu, ScfMenuItem ctxMenuItem)
        {
            _previousPage = _currentPage;
            uint prevId = _previousPage == null ? 0 : _previousPage.Id;
            Content cnt = _endPoint.GetContent(ctxMenuItem.TargetId);
            SideShow.Scf.Contents.ScfContent t = cnt as SideShow.Scf.Contents.ScfContent;
            _currentPage = t ?? _currentPage;
            if (_currentPage == null)
            {
                ContentMissingEvent cme = new ContentMissingEvent(_application.Id, _endPoint.Id, ctxMenuItem.TargetId);
                _channel.PostEvent(cme);
            }
            uint menuItemId = menuItem == null ? 0 : menuItem.Id;
            ScfContextMenuEvent e = new ScfContextMenuEvent(_application.Id, _endPoint.Id, prevId, ctxMenuItem.TargetId, menuItemId, ctxMenu.Id, ctxMenuItem.Id);
            _channel.PostEvent(e);
            if (_previousPage != null && _previousPage.ScfDocument.Body is ScfMenu)
            {
                if (_backStack.Count == 0 || (_backStack.Peek() != _previousPage && _previousPage != _currentPage))
                {
                    _backStack.Push(_previousPage);
                }
            }
            UserControl ctrl = GetPage(_currentPage);
            _container.Controls.Clear();
            _container.Controls.Add(ctrl);
            ctrl.Focus();
            OnCurrentPageChanged();
        }

        public void Navigate(ScfButton btn)
        {
            _previousPage = _currentPage;
            uint prevId = _previousPage == null ? 0 : _previousPage.Id;
            SideShow.Scf.Contents.ScfContent t = (SideShow.Scf.Contents.ScfContent)_endPoint.GetContent(btn.TargetId);
            _currentPage = t ?? _currentPage;
            SideShow.Scf.ScfButton key = GetKey(btn.Key);
            if (btn.Key == ScfButtonKey.Home)
            {
                _currentPage = (SideShow.Scf.Contents.ScfContent) _endPoint.GetContent(1);
            }
            if (_currentPage == null)
            {
                ContentMissingEvent cme = new ContentMissingEvent(_application.Id, _endPoint.Id, btn.TargetId);
                _channel.PostEvent(cme);
            }
            ScfNavigationEvent e = new ScfNavigationEvent(_application.Id, _endPoint.Id, prevId, btn.TargetId, key);
            _channel.PostEvent(e);
            if (_previousPage != null && _previousPage.ScfDocument.Body is ScfMenu)
            {
                if (_backStack.Count == 0 || (_backStack.Peek() != _previousPage && _previousPage != _currentPage))
                {
                    _backStack.Push(_previousPage);
                }
            }
            UserControl ctrl = GetPage(_currentPage);
            _container.Controls.Clear();
            _container.Controls.Add(ctrl);
            ctrl.Focus();
            OnCurrentPageChanged();
        }

        private SideShow.Scf.ScfButton GetKey(ScfButtonKey key)
        {
            switch (key)
            {
                case ScfButtonKey.Back:
                    return SideShow.Scf.ScfButton.Back;
                case ScfButtonKey.Enter:
                    return SideShow.Scf.ScfButton.Select;
                case ScfButtonKey.Home:
                    return SideShow.Scf.ScfButton.Select;
                case ScfButtonKey.Left:
                    return SideShow.Scf.ScfButton.Left;
                case ScfButtonKey.Right:
                    return SideShow.Scf.ScfButton.Right;
            }
            throw new ArgumentException();
        }

        private UserControl GetPage(SideShow.Scf.Contents.ScfContent content)
        {
            ScfUserControl ctrl = null;
            if (content.ScfDocument.Body is ScfMenu)
            {
                ctrl = new ScfMenuControl(this);

            }
            else if (content.ScfDocument.Body is ScfContent)
            {
                ctrl = new ScfContentControl(this);
            }

            if (ctrl != null)
            {
                ctrl.Dock = DockStyle.Fill;
                ctrl.ScfObject = content.ScfDocument.Body;
                //_pages.Add(content, ctrl);
            }
            return ctrl;
        }
    }
}
